package hamza.dali.flutter_osm_plugin

import android.content.Context
import android.content.SharedPreferences
import android.graphics.*
import android.graphics.drawable.BitmapDrawable
import android.graphics.drawable.Drawable
import android.os.Bundle
import android.util.Log
import android.view.Gravity
import android.view.View
import android.view.ViewGroup.LayoutParams.MATCH_PARENT
import android.view.ViewGroup.LayoutParams.WRAP_CONTENT
import android.widget.FrameLayout
import android.widget.LinearLayout
import androidx.core.content.ContextCompat
import androidx.core.content.res.ResourcesCompat
import androidx.core.graphics.BlendModeColorFilterCompat
import androidx.core.graphics.BlendModeCompat
import androidx.core.graphics.drawable.toBitmap
import androidx.lifecycle.*
import androidx.preference.PreferenceManager
import com.squareup.picasso.Callback
import com.squareup.picasso.Picasso
import com.squareup.picasso.Target
import hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.Companion.CREATED
import hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.Companion.DESTROYED
import hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.Companion.PAUSED
import hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.Companion.STARTED
import hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.Companion.STOPPED
import hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.Companion.mapSnapShot
import hamza.dali.flutter_osm_plugin.utilities.*
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding.OnSaveInstanceStateListener
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.platform.PlatformView
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers.Default
import kotlinx.coroutines.Dispatchers.Main
import kotlinx.coroutines.Job
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import org.osmdroid.bonuspack.routing.OSRMRoadManager
import org.osmdroid.bonuspack.routing.RoadManager
import org.osmdroid.bonuspack.utils.PolylineEncoder
import org.osmdroid.config.Configuration
import org.osmdroid.events.MapEventsReceiver
import org.osmdroid.events.MapListener
import org.osmdroid.events.ScrollEvent
import org.osmdroid.events.ZoomEvent
import org.osmdroid.tileprovider.tilesource.TileSourceFactory.MAPNIK
import org.osmdroid.util.BoundingBox
import org.osmdroid.util.GeoPoint
import org.osmdroid.views.CustomZoomButtonsController
import org.osmdroid.views.MapView
import org.osmdroid.views.overlay.*
import org.osmdroid.views.overlay.mylocation.GpsMyLocationProvider
import org.osmdroid.views.overlay.mylocation.MyLocationNewOverlay
import java.io.ByteArrayOutputStream
import kotlin.collections.component1
import kotlin.collections.component2
import kotlin.collections.set


fun FlutterOsmView.configZoomMap(call: MethodCall, result: MethodChannel.Result) {
    val args = call.arguments as HashMap<String, Any>
    this.map?.minZoomLevel = (args["minZoomLevel"] as Double)
    this.map?.maxZoomLevel = (args["maxZoomLevel"] as Double)
    stepZoom = args["stepZoom"] as Double
    initZoom = args["initZoom"] as Double


    result.success(200)
}

fun FlutterOsmView.getZoom(result: MethodChannel.Result) {
    try {
        result.success(this.map!!.zoomLevelDouble)
    } catch (e: Exception) {
        result.error("404", e.stackTraceToString(), null)
    }

}

class FlutterOsmView(
    private val context: Context,
    private val binaryMessenger: BinaryMessenger,
    private val id: Int,//viewId
    private val providerLifecycle: ProviderLifecycle,
    private var mapView: MapView
) :
    LifecycleObserver,
    OnSaveInstanceStateListener,
    PlatformView,
    MethodCallHandler {

    internal var map: MapView? = null
    private var locationNewOverlay: MyLocationNewOverlay? = null
    private var customMarkerIcon: Bitmap? = null
    private var customPersonMarkerIcon: Bitmap? = null
    private var customArrowMarkerIcon: Bitmap? = null
    private var customPickerMarkerIcon: Bitmap? = null
    private var staticMarkerIcon: HashMap<String, Bitmap> = HashMap()
    private val customRoadMarkerIcon = HashMap<String, Bitmap>()
    private val staticPoints: HashMap<String, MutableList<GeoPoint>> = HashMap()
    private var homeMarker: FlutterMarker? = null
    private val folderStaticPosition: FolderOverlay by lazy {
        FolderOverlay()
    }
    private val folderShape: FolderOverlay by lazy {
        FolderOverlay().apply {
            name = Constants.shapesNames
        }
    }
    private val folderCircles: FolderOverlay by lazy {
        FolderOverlay().apply {
            name = Constants.circlesNames
        }
    }
    private val folderRect: FolderOverlay by lazy {
        FolderOverlay().apply {
            name = Constants.regionNames
        }
    }
    private val folderRoad: FolderOverlay by lazy {
        FolderOverlay().apply {
            this.name = Constants.roadName
        }
    }
    private val folderMarkers: FolderOverlay by lazy {
        FolderOverlay().apply {
            this.name = Constants.markerNameOverlay
        }
    }

    private var flutterRoad: FlutterRoad? = null
    private var job: Job? = null
    private var jobFlow: Job? = null
    private var scope: CoroutineScope? = null


    private lateinit var methodChannel: MethodChannel


    private val provider: GpsMyLocationProvider by lazy {
        GpsMyLocationProvider(context)
    }

    private var mapEventsOverlay: MapEventsOverlay? = null


    private var roadManager: OSRMRoadManager? = null
    private var roadColor: Int? = null
    internal var stepZoom = Constants.stepZoom
    internal var initZoom = 10.0
    private var isTracking = false
    private var isEnabled = false
    private var visibilityInfoWindow = false

    companion object {
        val boundingWorldBox: BoundingBox =
            BoundingBox(
                85.0,
                180.0,
                -85.0,
                -180.0,
            )

    }

    private val staticOverlayListener by lazy {
        MapEventsOverlay(object : MapEventsReceiver {
            override fun singleTapConfirmedHelper(p: GeoPoint?): Boolean {

                methodChannel.invokeMethod("receiveSinglePress", p!!.toHashMap())

                return true
            }

            override fun longPressHelper(p: GeoPoint?): Boolean {

                methodChannel.invokeMethod("receiveLongPress", p!!.toHashMap())

                return true

            }

        })
    }


    private val mapListener by lazy {
        object : MapListener {
            override fun onScroll(event: ScrollEvent?): Boolean {
                return true
            }

            override fun onZoom(event: ZoomEvent?): Boolean {
                if (event!!.zoomLevel < Constants.zoomStaticPosition) {
                    val rect = Rect()
                    map?.getDrawingRect(rect)
                    map?.overlays?.remove(folderStaticPosition)
                } else if (markerSelectionPicker == null) {
                    if (map != null && !map!!.overlays.contains(folderStaticPosition)) {
                        map!!.overlays.add(folderStaticPosition)
                    }
                }
                return true
            }
        }
    }


    private var mainLinearLayout: FrameLayout = FrameLayout(context).apply {
        this.layoutParams =
            FrameLayout.LayoutParams(FrameLayout.LayoutParams(MATCH_PARENT, MATCH_PARENT))
    }
    private var markerSelectionPicker: FlutterPickerViewOverlay? = null

    init {
        providerLifecycle.getLifecyle()?.addObserver(this)

    }


    private fun initMap() {


        map = mapView

        map!!.layoutParams = MapView.LayoutParams(
            LinearLayout.LayoutParams(MATCH_PARENT, MATCH_PARENT)
        )
        map!!.isTilesScaledToDpi = true
        map!!.setMultiTouchControls(true)
        map!!.setTileSource(MAPNIK)
        map!!.isVerticalMapRepetitionEnabled = false
        map!!.isHorizontalMapRepetitionEnabled = false
        map!!.setScrollableAreaLimitDouble(mapSnapShot.boundingWorld())
        map!!.setScrollableAreaLimitLatitude(
            MapView.getTileSystem().maxLatitude,
            MapView.getTileSystem().minLatitude,
            0
        )
        map!!.zoomController.setVisibility(CustomZoomButtonsController.Visibility.NEVER)
        //
        map!!.minZoomLevel = 2.0
        map!!.controller.setZoom(mapSnapShot.zoomLevel(2.0))
        map!!.controller.setCenter(mapSnapShot.centerGeoPoint() ?: GeoPoint(0.0, 0.0))


        map!!.addMapListener(mapListener)
        map!!.overlayManager.add(0, staticOverlayListener)
        map!!.overlayManager.add(folderMarkers)

        mainLinearLayout.addView(map)


    }


    override fun onMethodCall(call: MethodCall, result: MethodChannel.Result) {
        try {
            when (call.method) {
                "map#setCache" -> {
                    setCacheMap()
                    result.success(null)
                }
                "map#clearCache#view" -> {
                    mapSnapShot.reset(all = true)
                    result.success(null)
                }
                "map#saveCache#view" -> {
                    saveCacheMap()
                    Log.e("map center", mapSnapShot.centerGeoPoint()!!.toHashMap().toString())
                    result.success(null)
                }
                "use#visiblityInfoWindow" -> {
                    visibilityInfoWindow = call.arguments as Boolean
                    result.success(null)
                }
                "config#Zoom" -> {
                    configZoomMap(call = call, result = result)
                }
                "Zoom" -> {
                    setZoom(call, result)
                }
                "get#Zoom" -> {
                    getZoom(result)
                }
                "change#stepZoom" -> {
                    stepZoom = call.arguments as Double
                    result.success(null)
                }
                "currentLocation" -> {
                    enableMyLocation()
                    result.success(isEnabled)
                }

                "showZoomController" -> {
                    val isZoomControllerVisible = call.arguments as Boolean
                    val visibility = if (isZoomControllerVisible) {
                        CustomZoomButtonsController.Visibility.SHOW_AND_FADEOUT
                    } else
                        CustomZoomButtonsController.Visibility.NEVER
                    map?.zoomController?.setVisibility(visibility)
                    result.success(null)
                }

                "initMap" -> {
                    initPosition(call, result)
                }
                "limitArea" -> {
                    limitCameraArea(call, result)
                }
                "remove#limitArea" -> {
                    removeLimitCameraArea(result)

                }
                "changePosition" -> {
                    changePosition(call, result)
                }
                "trackMe" -> {
                    trackUserLocation(result)
                }
                "deactivateTrackMe" -> {
                    deactivateTrackMe(call, result)
                }
                "map#center" -> {
                    result.success((map?.mapCenter as GeoPoint).toHashMap())
                }
                "user#position" -> {
                    if (locationNewOverlay == null) {
                        locationNewOverlay = MyLocationNewOverlay(provider, map)
                    }
                    locationNewOverlay?.let {
                        if (!it.isMyLocationEnabled) {
                            it.enableMyLocation()
                        }
                        currentUserPosition(call, result)
                    } ?: result.error("400", "Opps!error locationOverlay is NULL", "")
                }

                "user#pickPosition" -> {
                    pickPosition(call, result)
                }
                "goto#position" -> {
                    goToSpecificPosition(call, result)
                }
                "user#removeMarkerPosition" -> {
                    removePosition(call, result)
                }
                "user#removeroad" -> {
                    if (folderRoad.items.isNotEmpty()) {
                        mapSnapShot.clearCachedRoad()
                        folderRoad.items.clear()
                        map?.invalidate()
                    }
                    result.success(null)

                }
                "road" -> {
                    drawRoad(call, result)
                }
                "marker#icon" -> {
                    changeIcon(call, result)
                }
                "road#color" -> {
                    setRoadColor(call, result)
                }
                "drawRoad#manually" -> {
                    drawRoadManually(call, result)
                }
                "road#markers" -> {
                    setRoadMaker(call, result)
                }
                "staticPosition" -> {
                    staticPosition(call, result)
                }
                "staticPosition#IconMarker" -> {
                    staticPositionIconMaker(call, result)
                }
                "draw#circle" -> {
                    drawCircle(call, result)
                }
                "remove#circle" -> {
                    removeCircle(call, result)
                }
                "draw#rect" -> {
                    drawRect(call, result)
                }
                "remove#rect" -> {
                    removeRect(call, result)
                }
                "clear#shapes" -> {
                    folderCircles.items.clear()
                    folderRect.items.clear()
                    map?.invalidate()
                    result.success(null)

                }
                "advancedPicker#marker#icon" -> {
                    setCustomAdvancedPickerMarker(
                        call = call,
                        result = result,
                    )
                }
                "advanced#selection" -> {
                    startAdvancedSelection()
                    result.success(null)
                }
                "get#position#advanced#selection" -> {
                    confirmAdvancedSelection(result)
                }
                "confirm#advanced#selection" -> {
                    confirmAdvancedSelection(result, isFinished = true)
                }

                "cancel#advanced#selection" -> {
                    cancelAdvancedSelection()
                    result.success(null)
                }
                "map#orientation" -> {
                    mapOrientation(call, result)
                }
                "user#locationMarkers" -> {
                    changeLocationMarkers(call, result)
                }
                "add#Marker" -> {
                    addMarkerManually(call, result)
                }
                else -> {
                    result.notImplemented()
                }
            }

        } catch (e: Exception) {
            Log.e(e.cause.toString(), e.stackTraceToString())
            result.error("404", e.message, e.stackTraceToString())
        }
    }

    private fun setCacheMap() {
        if (mapSnapShot.centerGeoPoint() != null &&
            !mapSnapShot.centerGeoPoint()!!.eq(GeoPoint(0.0, 0.0))
        ) {

            map!!.controller.setCenter(mapSnapShot.centerGeoPoint())
            map!!.controller.setZoom(mapSnapShot.zoomLevel(initZoom))
            scope?.launch {
                mapSnapShot.markers().forEach { (point, bytes) ->
                    val icon = bytes?.let { getBitmap(bytes = it) }
                    val drawable = getDefaultIconDrawable(icon = icon, color = null)
                    withContext(Main) {
                        addMarker(
                            point,
                            dynamicMarkerBitmap = drawable,
                            animateTo = false,
                            zoom = mapSnapShot.zoomLevel(initZoom)
                        )
                    }
                }
            }

        }
        if (mapSnapShot.staticGeoPointsIcons().isNotEmpty()) {
            scope?.launch {
                mapSnapShot.staticGeoPointsIcons().forEach { (key, icon) ->
                    staticMarkerIcon[key] = getBitmap(icon)
                }
            }
        }

        if (mapSnapShot.staticGeoPoints().isNotEmpty()) {
            /**/
            scope?.launch {
                withContext(Default) {
                    mapSnapShot.staticGeoPointsIcons().forEach { (key, icon) ->
                        staticMarkerIcon[key] = getBitmap(icon)
                    }
                }
                mapSnapShot.staticGeoPoints().forEach { staticPoint ->
                    staticPoints[staticPoint.key] = staticPoint.value.first.toMutableList()
//                    val bitmapIcon = when (staticPoint.value.third != null) {
//                        true -> getBitmap(staticPoint.value.third!!)
//                        false -> getDefaultIconDrawable(icon = null, color = null).toBitmap()
//                    }
//                    staticMarkerIcon[staticPoint.key] = bitmapIcon
                    withContext(Main) {
                        showStaticPosition(
                            staticPoint.key,
                            staticPoint.value.second.toList()
                        )
                    }
                }
            }
        }
        mapSnapShot.lastCachedRoad()?.let { lastRoad ->
            if (lastRoad.roadPoints.isNotEmpty()) {
                if (!map!!.overlayManager.contains(folderRoad)) {
                    map!!.overlayManager.add(folderRoad)
                }
                val polyLine = Polyline(map!!)
                polyLine.setPoints(lastRoad.roadPoints)
                //customRoadMarkerIcon.p
                createRoad(
                    polyLine = polyLine,
                    colorRoad = lastRoad.roadColor,
                    roadWidth = lastRoad.roadWith,
                    listInterestPoints = lastRoad.listInterestPoints,
                    showPoiMarker = lastRoad.showIcons,
                )

                map!!.invalidate()
            }

        }

        when (mapSnapShot.advancedPicker()) {
            true -> startAdvancedSelection()
            false -> {
                isTracking = mapSnapShot.trackMyLocation()
                isEnabled = mapSnapShot.getEnableMyLocation()
                when (isEnabled || isTracking) {
                    true -> {
                        if (isEnabled) {
                            enableMyLocation()
                        }
                        if (isTracking) {
                            locationNewOverlay?.let { locationOverlay ->
                                when {
                                    !locationOverlay.isFollowLocationEnabled -> {
                                        locationOverlay.enableFollowLocation()
                                        onChangedLocation(locationOverlay)
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        clearCacheMap()
        methodChannel.invokeMethod("map#restored", null)
    }

    private fun saveCacheMap() {
        mapSnapShot.cache(
            geoPoint = map!!.mapCenter as GeoPoint,
            zoom = map!!.zoomLevelDouble,
            customArrowMarkerIcon = getBytesFromBitmap(customArrowMarkerIcon),
            customPersonMarkerIcon = getBytesFromBitmap(customPersonMarkerIcon),
            customPickerMarkerIcon = getBytesFromBitmap(customPickerMarkerIcon),
            customRoadMarkerIcon = HashMap(
                customRoadMarkerIcon.mapValues { m ->
                    getBytesFromBitmap(m.value)!!
                }
            )
        )
    }

    private fun clearCacheMap() {
        mapSnapShot.reset()
    }

    private fun setZoom(methodCall: MethodCall, result: MethodChannel.Result) {
        val args = methodCall.arguments as HashMap<String, Any>
        when (args.containsKey("stepZoom")) {
            true -> {
                var zoomInput = args["stepZoom"] as Double
                if (zoomInput == 0.0) {
                    zoomInput = stepZoom
                } else if (zoomInput == -1.0) {
                    zoomInput = -stepZoom
                }
                val zoom = map!!.zoomLevelDouble + zoomInput
                map!!.controller.setZoom(zoom)
            }
            false -> {
                if (args.containsKey("zoomLevel")) {
                    val level = args["zoomLevel"] as Double
                    map!!.controller.setZoom(level)
                }

            }
        }

        result.success(null)
    }

    private fun initPosition(methodCall: MethodCall, result: MethodChannel.Result) {
        @Suppress("UNCHECKED_CAST")
        val args = methodCall.arguments!! as HashMap<String, Double>
        val geoPoint = GeoPoint(args["lat"]!!, args["lon"]!!)
        val zoom = initZoom
        map!!.controller.setZoom(mapSnapShot.zoomLevel(zoom))
        map!!.controller.setCenter(mapSnapShot.centerGeoPoint() ?: geoPoint)
        methodChannel.invokeMethod("map#init", true)
        scope?.launch {
            mapSnapShot.cacheLocation(geoPoint, zoom)
        }
        result.success(null)
    }

    private fun changePosition(methodCall: MethodCall, result: MethodChannel.Result) {
        @Suppress("UNCHECKED_CAST")
        val args = methodCall.arguments!! as HashMap<String, Double>
        if (homeMarker != null) {
            map!!.overlays.remove(homeMarker)
        }
        //map!!.overlays.clear()
        val geoPoint = GeoPoint(args["lat"]!!, args["lon"]!!)
        val zoom = when (map!!.zoomLevelDouble) {
            0.0 -> initZoom
            else -> map!!.zoomLevelDouble
        }
        homeMarker = addMarker(geoPoint, zoom, null)

        result.success(null)
    }

    private fun addMarker(
        geoPoint: GeoPoint,
        zoom: Double,
        color: Int? = null,
        dynamicMarkerBitmap: Drawable? = null,
        imageURL: String? = null,
        animateTo: Boolean = true,
    ): FlutterMarker {
        map!!.controller.setZoom(zoom)
        if (animateTo)
            map!!.controller.animateTo(geoPoint)
        val marker: FlutterMarker = createMarker(geoPoint, color) as FlutterMarker
        marker.onClickListener = Marker.OnMarkerClickListener { marker, _ ->
            val hashMap = HashMap<String, Double>()
            hashMap["lon"] = marker!!.position.longitude
            hashMap["lat"] = marker.position.latitude
            methodChannel.invokeMethod("receiveGeoPoint", hashMap)
            true
        }
        when {
            dynamicMarkerBitmap != null -> {
                marker.icon = dynamicMarkerBitmap
                folderMarkers.items.add(marker)

            }
            imageURL != null && imageURL.isNotEmpty() -> {
                Picasso.get()
                    .load(imageURL)
                    .fetch(object : Callback {
                        override fun onSuccess() {
                            Picasso.get()
                                .load(imageURL)
                                .into(object : Target {
                                    override fun onBitmapLoaded(
                                        bitmapMarker: Bitmap?,
                                        from: Picasso.LoadedFrom?
                                    ) {

                                        marker.icon =
                                            BitmapDrawable(context.resources, bitmapMarker)
                                        map!!.overlays.add(marker)

                                    }

                                    override fun onBitmapFailed(
                                        e: java.lang.Exception?,
                                        errorDrawable: Drawable?
                                    ) {
                                        marker.icon = ContextCompat.getDrawable(
                                            context!!,
                                            R.drawable.ic_location_on_red_24dp
                                        )
                                        map!!.overlays.add(marker)

                                    }

                                    override fun onPrepareLoad(placeHolderDrawable: Drawable?) {
                                        // marker.icon = ContextCompat.getDrawable(context!!, R.drawable.ic_location_on_red_24dp)
                                    }

                                })
                        }

                        override fun onError(e: java.lang.Exception?) {
                            Log.e("error image", e?.stackTraceToString() ?: "")
                        }

                    })


            }
            else -> folderMarkers.items.add(marker)

        }

        map?.invalidate()
        return marker
    }

    private fun createMarker(geoPoint: GeoPoint, color: Int?, icon: Bitmap? = null): Marker {
        val marker = FlutterMarker(context, map!!, geoPoint)
        marker.visibilityInfoWindow(visibilityInfoWindow)
//        marker.longPress = object : LongClickHandler {
//            override fun invoke(marker: Marker): Boolean {
//                map!!.overlays.remove(marker)
//                map!!.invalidate()
//                return true
//            }
//        }
        val iconDrawable: Drawable = getDefaultIconDrawable(color, icon = icon)
        //marker.setPosition(geoPoint);
        marker.icon = iconDrawable
        //marker.setInfoWindow(new FlutterInfoWindow(creatWindowInfoView(),map!!,geoPoint));
        marker.position = geoPoint
        return marker
    }

    private fun getDefaultIconDrawable(color: Int?, icon: Bitmap? = null): Drawable {
        val iconDrawable: Drawable
        if (icon != null) {
            iconDrawable = BitmapDrawable(context.resources, icon)
            if (color != null) iconDrawable.setColorFilter(
                BlendModeColorFilterCompat.createBlendModeColorFilterCompat(
                    color,
                    BlendModeCompat.SRC_OVER
                )
            )
        } else if (customMarkerIcon != null) {
            iconDrawable = BitmapDrawable(context.resources, customMarkerIcon)
            if (color != null) iconDrawable.setColorFilter(
                BlendModeColorFilterCompat.createBlendModeColorFilterCompat(
                    color,
                    BlendModeCompat.SRC_OVER
                )
            )
        } else {
            iconDrawable =
                ContextCompat.getDrawable(context, R.drawable.ic_location_on_red_24dp)!!
        }
        return iconDrawable
    }

    private fun enableMyLocation() {

        if (markerSelectionPicker != null) {
            mainLinearLayout.removeView(markerSelectionPicker)
            if (!map!!.overlays.contains(folderShape))
                map!!.overlays.add(folderShape)
            if (!map!!.overlays.contains(folderRoad))
                map!!.overlays.add(folderRoad)
            if (!map!!.overlays.contains(folderStaticPosition))
                map!!.overlays.add(folderStaticPosition)
            markerSelectionPicker = null
        }

        if (locationNewOverlay == null) {
            locationNewOverlay = MyLocationNewOverlay(provider, map)
        }
        //locationNewOverlay!!.setPersonIcon()
        when (customPersonMarkerIcon != null) {
            true -> {
                when (customArrowMarkerIcon != null) {
                    true -> {
                        locationNewOverlay!!.setDirectionArrow(
                            customPersonMarkerIcon,
                            customArrowMarkerIcon
                        )
                        val mScale = map!!.context.resources.displayMetrics.density

                        locationNewOverlay!!.setPersonHotspot(
                            mScale * (customPersonMarkerIcon!!.width / 4f) + 0.5f,
                            mScale * (customPersonMarkerIcon!!.width / 3f) + 0.5f,
                        )
                    }
                    false -> locationNewOverlay!!.setPersonIcon(customPersonMarkerIcon)
                }

            }
        }
        locationNewOverlay?.let { location ->
            if (!location.isMyLocationEnabled) {
                isEnabled = true
                location.enableMyLocation()
            }
            mapSnapShot.setEnableMyLocation(isEnabled)
            location.runOnFirstFix {
                scope!!.launch(Main) {
                    val currentPosition = GeoPoint(location.lastFix)
                    map!!.controller.animateTo(currentPosition)
                }
            }
        }
        if (!map!!.overlays.contains(locationNewOverlay)) {
            map!!.overlays.add(locationNewOverlay)
        }

    }

    private fun onChangedLocation(locationOverlay: MyLocationNewOverlay) {
        //
        provider.startLocationProvider { location, source ->
            locationOverlay.onLocationChanged(location, source)
            val geoPMap = GeoPoint(location).toHashMap()
            methodChannel.invokeMethod("receiveUserLocation", geoPMap)

            //eventLocationSink?.success(geoPMap)
        }
    }


    private fun addMarkerManually(call: MethodCall, result: MethodChannel.Result) {
        val args = call.arguments as HashMap<*, *>
        val point = (args["point"] as HashMap<String, Double>).toGeoPoint()

        var bitmap = customMarkerIcon
        if (args.containsKey("icon")) {
            bitmap = getBitmap(args["icon"] as ByteArray)
            scope?.launch {
                mapSnapShot.overlaySnapShotMarker(
                    point = point,
                    icon = args["icon"] as ByteArray
                )
            }
        }


        addMarker(
            point,
            dynamicMarkerBitmap = getDefaultIconDrawable(null, icon = bitmap),
            zoom = map!!.zoomLevelDouble,
            animateTo = false
        )

        //map!!.overlays.add(marker)
        result.success(null)

    }

    private fun changeLocationMarkers(call: MethodCall, result: MethodChannel.Result) {
        val args: HashMap<String, Any> = call.arguments as HashMap<String, Any>
        try {
            customPersonMarkerIcon = getBitmap((args["personIcon"] as ByteArray))
            customArrowMarkerIcon = getBitmap((args["arrowDirectionIcon"] as ByteArray))
            result.success(null)
        } catch (e: Exception) {
            e.printStackTrace()
            result.success(e.message)

        }
    }

    private fun removeLimitCameraArea(result: MethodChannel.Result) {
        map!!.setScrollableAreaLimitDouble(boundingWorldBox)
        mapSnapShot.setBoundingWorld(boundingWorldBox)
        result.success(200)
    }

    private fun limitCameraArea(call: MethodCall, result: MethodChannel.Result) {
        val list = call.arguments as List<Double>
        val box = BoundingBox(
            list[0], list[1], list[2], list[3]
        )
        map!!.setScrollableAreaLimitDouble(box)
        mapSnapShot.setBoundingWorld(
            box = box
        )
        result.success(200)
    }

    private fun mapOrientation(call: MethodCall, result: MethodChannel.Result) {
        //map!!.mapOrientation = (call.arguments as Double?)?.toFloat() ?: 0f
        map!!.controller.animateTo(
            map!!.mapCenter,
            map!!.zoomLevelDouble,
            null,
            (call.arguments as Double?)?.toFloat() ?: 0f
        )
        map!!.invalidate()
        result.success(null)
    }


    private fun drawRoadManually(call: MethodCall, result: MethodChannel.Result) {
        val args: HashMap<String, Any> = call.arguments as HashMap<String, Any>

        val encodedWayPoints = (args["road"] as String)
        val colorRoad = (args["roadColor"] as List<Int>)
        val color = Color.rgb(colorRoad.first(), colorRoad.last(), colorRoad[1])
        val widthRoad = (args["roadWidth"] as Double)

        if (!map!!.overlays.contains(folderRoad)) {
            map!!.overlays.add(folderRoad)
        } else {
            folderRoad.items.clear()
        }

        val route = PolylineEncoder.decode(encodedWayPoints, 10, false)
        val polyLine = Polyline(map!!)
        polyLine.setPoints(route)
        polyLine.outlinePaint.color = color
        polyLine.outlinePaint.strokeWidth = widthRoad.toFloat()

        folderRoad.items.add(polyLine)

        mapSnapShot.cacheRoad(
            RoadSnapShot(
                roadPoints = route,
                roadColor = color,
                roadWith = widthRoad.toFloat(),
                showIcons = false,
            )
        )

        map!!.invalidate()
        result.success(null)
    }

    private fun trackUserLocation(result: MethodChannel.Result) {
        try {
            if (homeMarker != null) {
                folderMarkers.items.remove(homeMarker)
                map?.invalidate()
            }
            locationNewOverlay?.let { locationOverlay ->
                when {
                    !locationOverlay.isFollowLocationEnabled -> {
                        isTracking = true
                        locationOverlay.enableFollowLocation()
                        onChangedLocation(locationOverlay)
                        mapSnapShot.setTrackLocation(isTracking)
                        mapSnapShot.setEnableMyLocation(isEnabled)
                        result.success(true)
                    }
                    else -> result.success(null)

                }


            }
        } catch (e: Exception) {
            result.error("400", e.stackTraceToString(), "")
        }
    }

    private fun goToSpecificPosition(call: MethodCall, result: MethodChannel.Result) {
        val args = call.arguments!! as HashMap<String, *>
        val geoPoint = GeoPoint(args["lat"]!! as Double, args["lon"]!! as Double)
        //map!!.controller.zoomTo(defaultZoom)
        map!!.controller.animateTo(geoPoint)
        result.success(null)
    }


    private fun drawRect(call: MethodCall, result: MethodChannel.Result) {
        val args = call.arguments!! as HashMap<String, *>
        val geoPoint = GeoPoint(args["lat"]!! as Double, args["lon"]!! as Double)
        val key = args["key"] as String
        val colors = args["color"] as List<Double>
        val distance = (args["distance"] as Double)
        val stokeWidth = (args["stokeWidth"] as Double).toFloat()
        val color = Color.rgb(colors[0].toInt(), colors[1].toInt(), colors[2].toInt())

        val region: List<GeoPoint> =
            Polygon.pointsAsRect(geoPoint, distance, distance).toList() as List<GeoPoint>
        val p = Polygon(map!!)
        p.id = key
        p.points = region
        p.fillPaint.color = color
        p.fillPaint.style = Paint.Style.FILL
        p.fillPaint.alpha = 50
        p.outlinePaint.strokeWidth = stokeWidth
        p.outlinePaint.color = color
        p.setOnClickListener { polygon, _, _ ->
            polygon.closeInfoWindow()
            false
        }

        folderRect.items.removeAll {
            it is Polygon && it.id == key
        }
        folderRect.items.add(p)
        if (!map!!.overlays.contains(folderShape)) {
            map!!.overlays.add(folderShape)
            if (!folderShape.items.contains(folderRect)) {
                folderShape.add(folderRect)
            }
        }
        map!!.invalidate()
        result.success(null)
    }

    private fun removeRect(call: MethodCall, result: MethodChannel.Result) {
        val id = call.arguments as String?
        if (id != null)
            folderRect.items.removeAll {
                (it as Polygon).id == id
            }
        else {
            folderRect.items.clear()
        }
        map!!.invalidate()
        result.success(null)
    }

    private fun confirmAdvancedSelection(
        result: MethodChannel.Result,
        isFinished: Boolean = false
    ) {
        if (markerSelectionPicker != null) {
            //markerSelectionPicker!!.callOnClick()
            mainLinearLayout.removeView(markerSelectionPicker)
            val position = map!!.mapCenter as GeoPoint
            if (isFinished) {
                homeMarker = addMarker(position, map!!.zoomLevelDouble, null)
                markerSelectionPicker = null
                map!!.overlays.add(folderShape)
                map!!.overlays.add(folderRoad)
                map!!.overlays.add(folderStaticPosition)
                map!!.overlays.add(folderMarkers)
                map?.overlays?.add(0, staticOverlayListener)
                map?.invalidate()
                mapSnapShot.setAdvancedPicker(false)

                if (isTracking) {
                    isTracking = false
                    isEnabled = false
                }
            }
            result.success(position.toHashMap())

        }

    }

    private fun cancelAdvancedSelection() {
        if (markerSelectionPicker != null) {
            mainLinearLayout.removeView(markerSelectionPicker)
            if (isTracking) {
                try {
                    if (isEnabled) {
                        enableMyLocation()
                    }
                    locationNewOverlay?.let { locationOverlay ->
                        if (!locationOverlay.isFollowLocationEnabled) {
                            isTracking = true
                            locationOverlay.enableFollowLocation()
                            onChangedLocation(locationOverlay)
                        }
                    }
                } catch (e: Exception) {
                }
            }
            map!!.overlays.add(folderShape)
            map!!.overlays.add(folderRoad)
            map!!.overlays.add(folderStaticPosition)
            map!!.overlays.add(folderMarkers)
            map?.overlays?.add(0, staticOverlayListener)
            markerSelectionPicker = null
            mapSnapShot.setAdvancedPicker(false)

        }
    }

    private fun startAdvancedSelection() {
        map!!.overlays.clear()
        if (isTracking) {
            try {
                locationNewOverlay?.let { locationOverlay ->
                    if (locationOverlay.isFollowLocationEnabled) {
                        locationOverlay.disableFollowLocation()
                        locationOverlay.disableMyLocation()
                        provider.stopLocationProvider()
                    }
                }
            } catch (e: Exception) {
            }
        }
        map!!.invalidate()
        if (markerSelectionPicker != null) {
            mainLinearLayout.removeView(markerSelectionPicker)
        }
        val point = Point()
        map!!.projection.toPixels(map!!.mapCenter, point)
        val bitmap: Bitmap = customPickerMarkerIcon
            ?: ResourcesCompat.getDrawable(
                context.resources,
                R.drawable.ic_location_on_red_24dp,
                null
            )!!.toBitmap(
                64,
                64
            ) //BitmapFactory.decodeResource(, R.drawable.ic_location_on_red_24dp)?:customMarkerIcon

        markerSelectionPicker = FlutterPickerViewOverlay(
            bitmap, context, point, customPickerMarkerIcon != null
        )
        val params = FrameLayout.LayoutParams(
            WRAP_CONTENT,
            WRAP_CONTENT, Gravity.CENTER
        )
        markerSelectionPicker!!.layoutParams = params
        mainLinearLayout.addView(markerSelectionPicker)
        FlutterOsmPlugin.mapSnapShot.setAdvancedPicker(
            isActive = true
        )

    }

    private fun deactivateTrackMe(call: MethodCall, result: MethodChannel.Result) {
        isTracking = false
        isEnabled = false
        mapSnapShot.setTrackLocation(isTracking)
        mapSnapShot.setEnableMyLocation(isEnabled)
        try {
            locationNewOverlay?.let { locationOverlay ->
                when {
                    locationOverlay.isFollowLocationEnabled -> {
                        locationOverlay.disableFollowLocation()
                        locationOverlay.disableMyLocation()
                        provider.stopLocationProvider()
                    }
                    else -> result.success(null)
                }

            }
        } catch (e: Exception) {
            result.error("400", e.stackTraceToString(), "")
        }
        result.success(false)
    }

    private fun removeCircle(call: MethodCall, result: MethodChannel.Result) {
        val id = call.arguments as String?
        if (id != null)
            folderCircles.items.removeAll {
                (it as Polygon).id == id
            }
        else {
            folderCircles.items.clear()
        }
        map!!.invalidate()
        result.success(null)
    }

    private fun drawCircle(call: MethodCall, result: MethodChannel.Result) {
        val args = call.arguments!! as HashMap<String, *>
        val geoPoint = GeoPoint(args["lat"]!! as Double, args["lon"]!! as Double)
        val key = args["key"] as String
        val colors = args["color"] as List<Double>
        val radius = (args["radius"] as Double)
        val stokeWidth = (args["stokeWidth"] as Double).toFloat()
        val color = Color.rgb(colors[0].toInt(), colors[2].toInt(), colors[1].toInt())

        val circle: List<GeoPoint> = Polygon.pointsAsCircle(geoPoint, radius)
        val p = Polygon(map!!)
        p.id = key
        p.points = circle
        p.fillPaint.color = color
        p.fillPaint.style = Paint.Style.FILL
        p.fillPaint.alpha = 50
        p.outlinePaint.strokeWidth = stokeWidth
        p.outlinePaint.color = color
        p.setOnClickListener { polygon, _, _ ->
            polygon.closeInfoWindow()
            false
        }

        folderCircles.items.removeAll {
            it is Polygon && it.id == key
        }
        folderCircles.items.add(p)
        if (!map!!.overlays.contains(folderShape)) {
            map!!.overlays.add(folderShape)
            if (!folderShape.items.contains(folderCircles)) {
                folderShape.add(folderCircles)
            }
        }
        map!!.invalidate()
        result.success(null)
    }

    private fun drawRoad(call: MethodCall, result: MethodChannel.Result) {
        val args = call.arguments!! as HashMap<String, Any>

        val showPoiMarker = args["showMarker"] as Boolean
        val meanUrl = when (args["roadType"] as String) {
            "car" -> OSRMRoadManager.MEAN_BY_CAR
            "bike" -> OSRMRoadManager.MEAN_BY_BIKE
            "foot" -> OSRMRoadManager.MEAN_BY_FOOT
            else -> OSRMRoadManager.MEAN_BY_CAR
        }
        val listPointsArgs = args["wayPoints"] as List<HashMap<String, Double>>

        val listInterestPoints: List<GeoPoint> = when (args.containsKey("middlePoints")) {
            true -> args["middlePoints"] as List<HashMap<String, Double>>
            false -> emptyList()
        }.map { g ->
            GeoPoint(g["lat"]!!, g["lon"]!!)
        }.toList()

        val colorRoad: Int? = when (args.containsKey("roadColor")) {
            true -> {
                val colors = (args["roadColor"] as List<Int>)
                Color.rgb(colors.first(), colors.last(), colors[1])
            }
            else -> roadColor
        }
        val roadWidth: Float = when (args.containsKey("roadWidth")) {
            true -> (args["roadWidth"] as Double).toFloat()
            else -> 5f
        }
        flutterRoad?.road?.let {
            map!!.overlays.remove(it)
        }
        if (!map!!.overlays.contains(folderRoad)) {
            map!!.overlays.add(folderRoad)
        } else {
            folderRoad.items.clear()
        }
        map!!.invalidate()

        if (roadManager == null)
            roadManager = OSRMRoadManager(context, "json/application")
        roadManager?.let { manager ->
            manager.setMean(meanUrl)

            job = scope?.launch(Default) {
                val wayPoints = listPointsArgs.map {
                    GeoPoint(it["lat"]!!, it["lon"]!!)
                }.toList()
                withContext(Main) {
                    folderMarkers.items.removeAll {
                        (it is FlutterMarker && wayPoints.contains(it.position)) ||
                                (it is FlutterMarker && listInterestPoints.contains(it.position))
                    }
                    mapSnapShot.removeMarkersFromSnapShot(wayPoints)
                }
                val roadPoints = ArrayList(wayPoints)
                if (listInterestPoints.isNotEmpty()) {
                    roadPoints.addAll(1, listInterestPoints)
                }
                val road = manager.getRoad(roadPoints)
                withContext(Main) {
                    if (road.mRouteHigh.size > 2) {
                        val polyLine = RoadManager.buildRoadOverlay(road)
                        createRoad(
                            polyLine = polyLine,
                            colorRoad = colorRoad,
                            roadWidth = roadWidth,
                            showPoiMarker = showPoiMarker,
                            listInterestPoints = listInterestPoints,
                        )

                        mapSnapShot.cacheRoad(
                            RoadSnapShot(
                                roadPoints = road.mRouteHigh,
                                roadColor = colorRoad,
                                roadWith = roadWidth,
                                listInterestPoints = listInterestPoints,
                                showIcons = showPoiMarker
                            )
                        )

                        map!!.invalidate()
                    }
                    result.success(HashMap<String, Double>().apply {
                        this["duration"] = road.mDuration
                        this["distance"] = road.mLength
                    })
                }

            }
        }
    }

    private fun createRoad(
        polyLine: Polyline,
        colorRoad: Int?,
        showPoiMarker: Boolean,
        listInterestPoints: List<GeoPoint>,
        roadWidth: Float,
    ) {
        polyLine.setOnClickListener { _, _, eventPos ->
            methodChannel.invokeMethod("receiveSinglePress", eventPos?.toHashMap())
            true
        }
        /// set polyline color
        polyLine.outlinePaint.color = colorRoad ?: Color.GREEN


        flutterRoad = FlutterRoad(
            context,
            map!!,
            interestPoint = if (showPoiMarker) listInterestPoints else emptyList()
        )

        flutterRoad?.let { roadF ->
            roadF.markersIcons = customRoadMarkerIcon
            polyLine.outlinePaint.strokeWidth = roadWidth

            roadF.road = polyLine
            if (showPoiMarker) {
                // if (it.start != null)
                folderRoad.items.add(roadF.start.apply {
                    this.visibilityInfoWindow(visibilityInfoWindow)
                })
                //  if (it.end != null)
                folderRoad.items.add(roadF.end.apply {
                    this.visibilityInfoWindow(visibilityInfoWindow)
                })
                folderRoad.items.addAll(roadF.middlePoints)
            }

            folderRoad.items.add(roadF.road!!)
        }
    }

    private fun staticPositionIconMaker(call: MethodCall, result: MethodChannel.Result) {
        val hashMap: HashMap<String, Any> = call.arguments as HashMap<String, Any>

        try {
            val key = (hashMap["id"] as String)
            val bytes = (hashMap["bitmap"] as ByteArray)
            val bitmap = getBitmap(bytes)
            val refresh = hashMap["refresh"] as Boolean
            staticMarkerIcon[key] = bitmap
            scope?.launch {
                mapSnapShot.addToIconsStaticGeoPoints(key, bytes)
                if (staticPoints.containsKey(key) && refresh) {
                    showStaticPosition(
                        key,
                        mapSnapShot.staticGeoPoints()[key]!!.second
                    )
                }
            }
            result.success(null)
        } catch (e: java.lang.Exception) {
            Log.e("id", hashMap["id"].toString())
            Log.e("err static point marker", e.stackTraceToString())
            result.error("400", "error to getBitmap static Position", "")
            staticMarkerIcon = HashMap()
        }
    }

    private fun staticPosition(call: MethodCall, result: MethodChannel.Result) {
        val map = call.arguments as HashMap<String, Any>
        val id = map["id"] as String?
        val points = map["point"] as MutableList<HashMap<String, Double>>?
        val geoPoints: MutableList<GeoPoint> = emptyList<GeoPoint>().toMutableList()
        val angleGeoPoints: MutableList<Double> = emptyList<Double>().toMutableList()
        for (hashMap in points!!) {
            geoPoints.add(GeoPoint(hashMap["lat"]!!, hashMap["lon"]!!))
            when (hashMap.containsKey("angle")) {
                true -> angleGeoPoints.add(hashMap["angle"] ?: 0.0)
                else -> angleGeoPoints.add(0.0)
            }
        }
        if (staticPoints.containsKey(id)) {
            Log.e(id, "" + points.size)
            staticPoints[id]?.clear()
            staticPoints[id]?.addAll(geoPoints)
            if (folderStaticPosition.items.isNotEmpty())
                folderStaticPosition.remove(folderStaticPosition.items.first {
                    (it as FolderOverlay).name?.equals(id) == true
                })
        } else {
            staticPoints[id!!] = geoPoints
        }
        showStaticPosition(id!!, angleGeoPoints.toList())
        scope?.launch {
            mapSnapShot.addToStaticGeoPoints(
                id, Pair(
                    geoPoints.toList(),
                    angleGeoPoints.toList(),
                )
            )
        }
        result.success(null)
    }

    private fun setRoadMaker(call: MethodCall, result: MethodChannel.Result) {
        try {
            val hashMap = call.arguments!! as HashMap<String, ByteArray>
            hashMap.forEach { (key, bytes) ->
                customRoadMarkerIcon[key] = BitmapFactory.decodeByteArray(bytes, 0, bytes.size)
            }
            result.success(null)
        } catch (e: Exception) {
            Log.d("err", e.stackTraceToString())
            result.error("400", "Opss!Erreur", e.stackTraceToString())
        }
    }

    private fun setRoadColor(call: MethodCall, result: MethodChannel.Result) {
        val argb = call.arguments!! as List<Int>
        roadColor = Color.rgb(argb[0], argb[1], argb[2])
        result.success(null)
    }

    private fun changeIcon(call: MethodCall, result: MethodChannel.Result) {
        try {
            customMarkerIcon = getBitmap(call.arguments as ByteArray)
            //customMarkerIcon.recycle();
            result.success(null)
        } catch (e: Exception) {
            Log.d("err", e.stackTraceToString())
            customMarkerIcon = null
            result.error("500", "Cannot make markerIcon custom", "")
        }
    }

    private fun setCustomAdvancedPickerMarker(call: MethodCall, result: MethodChannel.Result) {
        try {
            customPickerMarkerIcon = getBitmap(call.arguments as ByteArray)
            //customMarkerIcon.recycle();
            result.success(null)
        } catch (e: Exception) {
            Log.d("err", e.stackTraceToString())
            customMarkerIcon = null
            result.error("500", "Cannot make markerIcon custom", "")
        }
    }

    private fun pickPosition(call: MethodCall, result: MethodChannel.Result) {
        //val usingCamera=call.arguments as Boolean

        val args = call.arguments as Map<String, Any>
        val marker: Drawable? = if (args.containsKey("icon")) {
            val bitmap = getBitmap(args["icon"] as ByteArray)
            BitmapDrawable(context.resources, bitmap)
        } else null
        val imageURL: String? = if (args.containsKey("imageURL")) {
            args["imageURL"] as String
        } else null

        if (mapEventsOverlay == null) {
            if (map!!.overlays.first() is MapEventsOverlay)
                map!!.overlays.removeFirst()
            mapEventsOverlay = MapEventsOverlay(object : MapEventsReceiver {
                override fun singleTapConfirmedHelper(p: GeoPoint?): Boolean {

                    val pMarker = addMarker(
                        p!!, map!!.zoomLevelDouble,
                        null,
                        marker,
                        imageURL,
                    )
                    scope?.launch {
                        mapSnapShot.overlaySnapShotMarker(
                            point = p,
                            icon = getBytesFromBitmap(pMarker.icon.toBitmap())!!
                        )
                    }
                    result.success(p.toHashMap())
                    if (mapEventsOverlay != null) {
                        mapEventsOverlay = null
                        map!!.overlays.removeFirst()
                        map!!.overlays.add(0, staticOverlayListener)
                    }
                    return true
                }

                override fun longPressHelper(p: GeoPoint?): Boolean {
                    return true
                }

            })
            if (mapEventsOverlay != null)
                map!!.overlays.add(0, mapEventsOverlay)
        }

    }

    private fun removePosition(call: MethodCall, result: MethodChannel.Result) {
        val geoMap = call.arguments as HashMap<String, Double>
        deleteMarker(geoMap.toGeoPoint())
        result.success(null)
    }

    private fun deleteMarker(geoPoint: GeoPoint) {
        val geoMarkers = folderMarkers.items.filterIsInstance<FlutterMarker>().filter { marker ->
            marker.position.eq(geoPoint)
        }
        if (geoMarkers.isNotEmpty()) {
            folderMarkers.items.removeAll(geoMarkers)
            scope?.launch {
                mapSnapShot.removeMarkersFromSnapShot(
                    removedPoints = geoMarkers.map {
                        it.position
                    }
                )
            }
            map!!.overlays.remove(folderMarkers)
            map!!.overlays.add(folderMarkers)
            map!!.invalidate()
        }

    }

    private fun currentUserPosition(call: MethodCall, result: MethodChannel.Result) {
        locationNewOverlay?.let { locationOverlay ->
            locationOverlay.runOnFirstFix {
                scope!!.launch(Main) {
                    locationOverlay.lastFix?.let { location ->
                        val point = GeoPoint(
                            location.latitude,
                            location.longitude,
                        )
                        locationOverlay.disableMyLocation()
                        result.success(point.toHashMap())
                    } ?: result.error("400", "we cannot get the current position!", "")
                }
            }

        } ?: result.error("400", "we cannot get the current position!", "")
    }

    private fun showStaticPosition(idStaticPosition: String, angles: List<Double> = emptyList()) {

        var overlay: FolderOverlay? = folderStaticPosition.items.firstOrNull {
            (it as FolderOverlay).name?.equals(idStaticPosition) == true
        } as FolderOverlay?

        overlay?.let {
            it.items.clear()
        }
        if (overlay != null) {
            folderStaticPosition.remove(overlay)
        }
        if (overlay == null) {
            overlay = FolderOverlay().apply {
                name = idStaticPosition
            }
        }

        staticPoints[idStaticPosition]?.forEachIndexed { index, geoPoint ->
            val marker = FlutterMarker(context, map!!)
            marker.position = geoPoint

            marker.defaultInfoWindow()
            marker.visibilityInfoWindow(visibilityInfoWindow)
            marker.onClickListener = Marker.OnMarkerClickListener { marker, _ ->
                val hashMap = HashMap<String, Double>()
                hashMap["lon"] = marker!!.position.longitude
                hashMap["lat"] = marker.position.latitude
                methodChannel.invokeMethod("receiveGeoPoint", hashMap)
                true
            }
            if (staticMarkerIcon.isNotEmpty() && staticMarkerIcon.containsKey(idStaticPosition)) {
                marker.setIconMaker(
                    null,
                    staticMarkerIcon[idStaticPosition],
                    angle = when (angles.isNotEmpty()) {
                        true -> angles[index]
                        else -> 0.0
                    }
                )
            } else {
                marker.setIconMaker(null, null)
            }
            overlay.add(marker)
        }
        folderStaticPosition.add(overlay)
        if (map!!.zoomLevelDouble > 10.0 && !mapSnapShot.advancedPicker()) {
            if (map!!.overlays.contains(folderStaticPosition)) {
                map!!.overlays.remove(folderStaticPosition)
            }
            if (!map!!.overlays.contains(folderStaticPosition)) {
                map!!.overlays.add(folderStaticPosition)

            }
            map!!.invalidate()
        }

    }


    private fun getBitmap(bytes: ByteArray): Bitmap {
        return BitmapFactory.decodeByteArray(bytes, 0, bytes.size)
    }

    private fun getBytesFromBitmap(bitmap: Bitmap?): ByteArray? {
        if (bitmap == null) {
            return null
        }
        val stream = ByteArrayOutputStream()
        bitmap.compress(Bitmap.CompressFormat.PNG, 90, stream)
        return stream.toByteArray()
    }


    override fun getView(): View {
        return mainLinearLayout
    }

    override fun dispose() {
        mainLinearLayout.removeAllViews()
        //clearCacheMap()
        //map!!.onDetach()
        // map = null
    }

    override fun onFlutterViewAttached(flutterView: View) {
        //   map!!.onAttachedToWindow()
        if (map == null) {
            val prefs: SharedPreferences = PreferenceManager.getDefaultSharedPreferences(context)
            Configuration.getInstance()
                .load(context, PreferenceManager.getDefaultSharedPreferences(context))
//            initMap()
//            map?.forceLayout()
        }

    }


    override fun onFlutterViewDetached() {
        map!!.onDetach()
        staticMarkerIcon.clear()
        staticPoints.clear()
        customMarkerIcon = null
        customRoadMarkerIcon.clear()
//        mainLinearLayout.removeAllViews()
//        map!!.onDetach()
//        map = null
    }


    override fun onSaveInstanceState(bundle: Bundle) {
        bundle.putString("center", "${map!!.mapCenter.latitude},${map!!.mapCenter.longitude}")
        bundle.putString("zoom", map!!.zoomLevelDouble.toString())
    }

    override fun onRestoreInstanceState(bundle: Bundle?) {
        Log.d("osm data", bundle?.getString("center") ?: "")
    }

    @OnLifecycleEvent(Lifecycle.Event.ON_CREATE)
    fun onCreate(owner: LifecycleOwner) {
        FlutterOsmPlugin.state.set(CREATED)
        methodChannel = MethodChannel(binaryMessenger, "plugins.dali.hamza/osmview_${id}")
        methodChannel.setMethodCallHandler(this)
        //eventChannel = EventChannel(binaryMessenger, "plugins.dali.hamza/osmview_stream_${id}")
        //eventChannel.setStreamHandler(this)
        //methodChannel.invokeMethod("map#init", true)

        scope = owner.lifecycle.coroutineScope
        folderStaticPosition.name = Constants.nameFolderStatic
        Configuration.getInstance().load(
            context,
            PreferenceManager.getDefaultSharedPreferences(context)
        )
        initMap()
        // map!!.forceLayout()
        Log.e("osm", "osm flutter plugin create")

    }

    @OnLifecycleEvent(Lifecycle.Event.ON_START)
    fun onStart(owner: LifecycleOwner) {
        FlutterOsmPlugin.state.set(STARTED)
        Log.e("osm", "osm flutter plugin start")
    }

    @OnLifecycleEvent(Lifecycle.Event.ON_RESUME)
    fun onResume(owner: LifecycleOwner) {
        FlutterOsmPlugin.state.set(FlutterOsmPlugin.RESUMED)
        Log.e("osm", "osm flutter plugin resume")
        if (map == null) {
            initMap()
        }
        map?.onResume()
        reStartFollowLocation()


    }

    @OnLifecycleEvent(Lifecycle.Event.ON_PAUSE)
    fun onPause(owner: LifecycleOwner) {
        FlutterOsmPlugin.state.set(PAUSED)
        stopFollowLocation()
        //val prefs: SharedPreferences = PreferenceManager.getDefaultSharedPreferences(context)
        //Configuration.getInstance().save(context, prefs)
        map?.onPause()
        Log.e("osm", "osm flutter plugin pause")

    }

    @OnLifecycleEvent(Lifecycle.Event.ON_STOP)
    fun onStop(owner: LifecycleOwner) {
        FlutterOsmPlugin.state.set(STOPPED)
        Log.e("osm", "osm flutter plugin stopped")
        job?.let {
            if (it.isActive) {
                it.cancel()
            }
        }
        jobFlow?.let {
            if (it.isActive) {
                it.cancel()
            }
        }
        jobFlow = null
        job = null

    }

    @OnLifecycleEvent(Lifecycle.Event.ON_DESTROY)
    fun onDestroy(owner: LifecycleOwner) {
        mainLinearLayout.removeAllViews()

        mapSnapShot.reset(all = true)

        //map!!.onDetach()
        methodChannel.setMethodCallHandler(null)
        //configuration!!.osmdroidTileCache.delete()
        //configuration = null
        //eventChannel.setStreamHandler(null)
        map = null
        FlutterOsmPlugin.state.set(DESTROYED)

    }

    private fun reStartFollowLocation() {
        if (isEnabled || isTracking) {

            if (locationNewOverlay == null) {
                locationNewOverlay = MyLocationNewOverlay(provider, map)
            }

            locationNewOverlay?.also { myLocation ->
                if (isEnabled) {
                    myLocation.enableMyLocation()
                }
                if (isTracking) {
                    myLocation.enableFollowLocation()
                    onChangedLocation(myLocation)

                }
            }
        } else {
            locationNewOverlay?.also { myLocation ->
                if (myLocation.isMyLocationEnabled)
                    myLocation.disableMyLocation()
                if (myLocation.isFollowLocationEnabled)
                    myLocation.disableFollowLocation()
            }
        }
    }

    private fun stopFollowLocation() {
        if (isTracking || isEnabled) {
            locationNewOverlay?.also { myLocation ->

                if (myLocation.isFollowLocationEnabled) {
                    myLocation.disableFollowLocation()

                }
                if (myLocation.isMyLocationEnabled) {
                    myLocation.disableMyLocation()
                    //provider.stopLocationProvider()
                }
                provider.stopLocationProvider()
            }
            locationNewOverlay = null
        }
    }
}

